"""
This script shows an example of how to use Niryo One's vision to
make a conditioning with any of the objects supplied with the Vision Kit
The script works in 2 ways:
- One where all the vision process is made on the robot
- One where the vision process is made on the computer

The first one shows how easy it is to use Niryo One's vision kit with Python TCP API
The second demonstrates a way to do image processing from user's computer. It highlights
the fact that the user can imagine every type of process on his computer.

The objects will be conditioned in a grid of dimension grid_dimension. If the grid is completed,
objects will be pack over the lower level
"""

from pyniryo import *

# -- MUST Change these variables
robot_ip_address = "10.10.10.10"  # IP address of the Niryo One
workspace_name = "workspace_1"  # Robot's Workspace Name

# -- Can Change these variables
grid_dimension = (3, 3)  # conditioning grid dimension
vision_process_on_robot = True  # boolean to indicate if the image processing append on the Robot
display_stream = True  # Only used if vision on computer

# -- Should Change these variables
# The pose from where the image processing happens
observation_pose = PoseObject(
    x=0.20, y=0., z=0.3,
    roll=0.0, pitch=1.57, yaw=0.0,
)

# Center of the conditioning area
center_conditioning_pose = PoseObject(
    x=0.0, y=-0.25, z=0.12,
    roll=-0., pitch=1.57, yaw=-1.57
)


# -- MAIN PROGRAM

def process(niryo_robot):
    # Initializing variables
    obj_pose = None
    try_without_success = 0
    count = 0
    if not vision_process_on_robot:
        mtx, dist = niryo_robot.get_camera_intrinsics()
    else:
        mtx = dist = None
    # Loop
    while try_without_success < 5:
        # Moving to observation pose
        niryo_robot.move_pose(observation_pose)

        if vision_process_on_robot:
            ret = niryo_robot.get_target_pose_from_cam(workspace_name,
                                                       height_offset=0.0,
                                                       shape=ObjectShape.ANY,
                                                       color=ObjectColor.ANY)
            obj_found, obj_pose, shape, color = ret

        else:  # Home made image processing
            img_compressed = niryo_robot.get_img_compressed()
            img = uncompress_image(img_compressed)
            img = undistort_image(img, mtx, dist)
            # extracting working area
            status, im_work = extract_img_workspace(img, workspace_ratio=1.0)
            if not status:
                print("Unable to find markers")
                try_without_success += 1
                if display_stream:
                    cv2.imshow("Last image saw", img)
                    cv2.waitKey(25)
                continue

            # Applying Threshold on ObjectColor
            color_hsv_setting = ColorHSV.ANY.value
            img_thresh = threshold_hsv(im_work, *color_hsv_setting)

            if display_stream:
                show_img("Last image saw", img, wait_ms=0)
                show_img("Image thresh", img_thresh, wait_ms=30)
            # Getting biggest contour/blob from threshold image
            contour = biggest_contour_finder(img_thresh)
            if contour is None or len(contour) == 0:
                print("No blob found")
                obj_found = False
            else:
                img_thresh_rgb = cv2.cvtColor(img_thresh, cv2.COLOR_GRAY2BGR)
                draw_contours(img_thresh_rgb, [contour])
                show_img("Image thresh", img_thresh, wait_ms=30)

                # Getting contour/blob center and angle
                cx, cy = get_contour_barycenter(contour)
                cx_rel, cy_rel = relative_pos_from_pixels(im_work, cx, cy)
                angle = get_contour_angle(contour)

                # Getting object world pose from relative pose
                obj_pose = niryo_robot.get_target_pose_from_rel(workspace_name,
                                                                height_offset=0.0,
                                                                x_rel=cx_rel, y_rel=cy_rel,
                                                                yaw_rel=angle)
                obj_found = True
        if not obj_found:
            try_without_success += 1
            continue
        # Everything is good, so we going to object
        niryo_robot.pick_from_pose(obj_pose)

        # Computing new place pose
        offset_x = count % grid_dimension[0] - grid_dimension[0] // 2
        offset_y = (count // grid_dimension[1]) % 3 - grid_dimension[1] // 2
        offset_z = count // (grid_dimension[0] * grid_dimension[1])
        place_pose = center_conditioning_pose.copy_with_offsets(0.05 * offset_x, 0.05 * offset_y, 0.025 * offset_z)

        # Placing
        niryo_robot.place_from_pose(place_pose)

        try_without_success = 0
        count += 1


if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Changing tool
    robot.update_tool()
    # Calibrate robot if robot needs calibration
    robot.calibrate_auto()
    # Launching main process
    process(robot)
    # Ending
    robot.go_to_sleep()
    # Releasing connection
    robot.close_connection()



import math
import numpy as np

from .enums import *
from .markers_detection import extract_img_markers, draw_markers

__all__ = [
    "threshold_hsv",
    "morphological_transformations",
    "get_contour_barycenter",
    "get_contour_angle",
    "biggest_contour_finder",
    "biggest_contours_finder",
    "draw_contours",
    "extract_img_workspace",
    "relative_pos_from_pixels",

    "show_img_and_check_close",
    "show_img",
    "show_img_and_wait_close",

    "compress_image",
    "uncompress_image",
    "add_annotation_to_image",
    "undistort_image",
    "resize_img",
    "concat_imgs",

    "extract_img_from_ros_msg",

    "debug_threshold_color",
    "debug_markers",
]


# Image Processing
def threshold_hsv(img, list_min_hsv, list_max_hsv, reverse_hue=False, use_s_prime=False):
    """
    Take BGR image (OpenCV imread result) and return thresholded image
    according to values on HSV (Hue, Saturation, Value)
    Pixel will worth 1 if a pixel has a value between min_v and max_v for all channels

    :param img: image BGR if rgb_space = False
    :type img: numpy.array
    :param list_min_hsv: list corresponding to [min_value_H,min_value_S,min_value_V]
    :type list_min_hsv: list[int]
    :param list_max_hsv: list corresponding to [max_value_H,max_value_S,max_value_V]
    :type list_max_hsv: list[int]
    :param use_s_prime: True if you want to use S channel as S' = S x V else classic
    :type use_s_prime: bool
    :param reverse_hue: Useful for Red color cause it is at both extremum
    :type reverse_hue: bool
    :return: threshold image
    :rtype: numpy.array
    """
    frame_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    if use_s_prime:
        frame_hsv[:, :, 1] = (1. / 255) * frame_hsv[:, :, 1] * frame_hsv[:, :, 2].astype(np.uint8)

    if not reverse_hue:
        return cv2.inRange(frame_hsv, tuple(list_min_hsv), tuple(list_max_hsv))
    else:
        list_min_v_c = list(list_min_hsv)
        list_max_v_c = list(list_max_hsv)
        lower_bound_red, higher_bound_red = sorted([list_min_v_c[0], list_max_v_c[0]])
        list_min_v_c[0], list_max_v_c[0] = 0, lower_bound_red
        low_red_im = cv2.inRange(frame_hsv, tuple(list_min_v_c), tuple(list_max_v_c))
        list_min_v_c[0], list_max_v_c[0] = higher_bound_red, 179
        high_red_im = cv2.inRange(frame_hsv, tuple(list_min_v_c), tuple(list_max_v_c))
        return cv2.addWeighted(low_red_im, 1.0, high_red_im, 1.0, 0)


def morphological_transformations(im_thresh, morpho_type=MorphoType.CLOSE, kernel_shape=(5, 5),
                                  kernel_type=KernelType.ELLIPSE):
    """
    Take black & white image and apply morphological transformation

    :param im_thresh: Black & White Image
    :type im_thresh: numpy.array
    :param morpho_type: CLOSE/OPEN/ERODE/DILATE => See on OpenCV/Google if you do not know these words
    :type morpho_type: MorphoType
    :param kernel_shape: tuple corresponding to the size of the kernel
    :type kernel_shape: tuple[float]
    :param kernel_type: RECT/ELLIPSE/CROSS => see on OpenCV
    :type kernel_type: KernelType
    :return: image after processing
    :rtype: numpy.array
    """
    if not isinstance(morpho_type, MorphoType) or not isinstance(kernel_type, KernelType):
        raise TypeError

    kernel = cv2.getStructuringElement(kernel_type.value, kernel_shape)

    return cv2.morphologyEx(im_thresh, morpho_type.value, kernel)


# Contours
def get_contour_barycenter(contour):
    """
    Return barycenter of an OpenCV Contour

    :param contour:
    :type contour: OpenCV Contour
    :return: Barycenter X, Barycenter Y
    :rtype: int, int
    """
    moments = cv2.moments(contour)
    if moments['m00'] == 0:
        rect = cv2.boundingRect(contour)
        cx, cy, _, _ = rect
    else:
        cx = int(moments['m10'] / moments['m00'])
        cy = int(moments['m01'] / moments['m00'])

    return cx, cy


def get_contour_angle(contour):
    """
    Return orientation of a contour according to the smallest side
    in order to be well oriented for gripper

    :param contour: contour
    :type contour: OpenCV Contour
    :return: Angle in radians
    :rtype: float
    """
    rotrect = cv2.minAreaRect(contour)
    angle = rotrect[-1]
    size1, size2 = rotrect[1][0], rotrect[1][1]
    ratio_size = float(size1) / float(size2)
    if 1.25 > ratio_size > 0.75:
        if angle < -45:
            angle = 90 + angle
    else:
        if size1 < size2:
            angle = angle + 180
        else:
            angle = angle + 90

        if angle > 90:
            angle = angle - 180

    return math.radians(angle)


def biggest_contour_finder(img):
    res = biggest_contours_finder(img, nb_contours_max=1)
    if not res:
        return res
    else:
        return res[0]


def biggest_contours_finder(img, nb_contours_max=3):
    """
    Function to find the biggest contour in an binary image

    :param img: Binary Image
    :type img: numpy.array
    :param nb_contours_max: maximal number of contours which will be returned
    :type nb_contours_max: int
    :return: biggest contours found
    :rtype: list[OpenCV Contour]
    """
    contours = cv2.findContours(img, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)[-2]
    if not contours:
        return []
    contours_area = list()
    for cnt in contours:
        contours_area.append(cv2.contourArea(cnt))
    biggest_contours = []
    le = len(contours_area)
    if nb_contours_max > le:
        nb_contours = le
        id_contours_sorted_init = list(range(nb_contours))
    else:
        nb_contours = nb_contours_max
        id_contours_sorted_init = np.argpartition(contours_area, -nb_contours)[-nb_contours:]
    id_contours_sorted = [x for x in sorted(id_contours_sorted_init, key=lambda idi: -contours_area[idi])]

    for i in range(nb_contours):
        id_used = id_contours_sorted[i]

        if contours_area[id_used] < 400:
            break

        biggest_contours.append(contours[id_used])
    return biggest_contours


def draw_contours(img, contours):
    """
    Draw a list of contour on an image and return the drawing image

    :param img: Image
    :type img: numpy.array
    :param contours: contours list
    :type contours: list[OpenCV Contour]
    :return: Image with drawing
    :rtype: numpy.array
    """
    if len(img.shape) == 2:
        img_bgr = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    else:
        img_bgr = img.copy()
    cv2.drawContours(img_bgr, contours, -1, (255, 0, 0), 3)
    return img_bgr


def extract_img_workspace(img, workspace_ratio=1.0):
    """
    Extract working area from an image thanks to 4 Niryo's markers

    :param img: OpenCV image which contain 4 Niryo's markers
    :type img: numpy.array
    :param workspace_ratio: Ratio between the width and the height of the area represented by the markers
    :type workspace_ratio: float
    :return: extracted and warped working area image
    :rtype: numpy.array
    """
    return extract_img_markers(img, workspace_ratio)


def relative_pos_from_pixels(img, x_pixels, y_pixels):
    """
    Transform a pixels position to a relative position

    :param img: Image where the object is detected
    :type img: numpy.array
    :param x_pixels: coordinate X
    :type x_pixels: int
    :param y_pixels: coordinate Y
    :type y_pixels: int
    :return: X relative, Y relative
    :rtype: float, float
    """
    return float(x_pixels) / img.shape[1], float(y_pixels) / img.shape[0]


# - SHOW FUNCTIONS
keys_leave = [27, ord('q')]


def show_img_and_check_close(window_name, img):
    """
    Display an image and check whether the user want to close

    :param window_name: window's name
    :type window_name: str
    :param img: Image
    :type img: numpy.array
    :return: boolean indicating if the user wanted to leave
    :rtype: bool
    """
    cv2.imshow(window_name, img)
    return cv2.waitKey(1) in keys_leave


def show_img(window_name, img, wait_ms=1):
    """
    Display an image during a certain time

    :param window_name: window's name
    :type window_name: str
    :param img: Image
    :type img: numpy.array
    :param wait_ms: Wait time in milliseconds
    :type wait_ms: int
    :return: value of the key pressed during the display
    :rtype: int
    """
    if type(wait_ms) == float:
        wait_ms = int(wait_ms)
    if wait_ms < 1:
        wait_ms = 1
    cv2.imshow(window_name, img)
    return cv2.waitKey(wait_ms)


def show_img_and_wait_close(window_name, img):
    """
    Display an image and wait that the user close it

    :param window_name: window's name
    :type window_name: str
    :param img: Image
    :type img: numpy.array
    :return: None
    :rtype: None
    """
    cv2.imshow(window_name, img)
    cv2.waitKey(0)


# -- IMAGE EDITING
def compress_image(img, quality=90):
    """
    Compress OpenCV image

    :param img: OpenCV Image
    :type img: numpy.array
    :param quality: integer between 1 - 100. The higher it is, the less information will be lost,
     but the heavier the compressed image will be
    :type quality: int
    :return: status & string representing compressed image
    :rtype: bool, str
    """
    result, encimg = cv2.imencode('.jpg', img, [int(cv2.IMWRITE_JPEG_QUALITY), quality])
    if not result:
        return False, None

    return True, np.array(encimg).tostring()


def uncompress_image(compressed_image):
    """
    Take a compressed img and return an OpenCV image

    :param compressed_image: compressed image
    :type compressed_image: str
    :return: OpenCV image
    :rtype: numpy.array
    """
    np_arr = np.fromstring(compressed_image, np.uint8)
    return cv2.imdecode(np_arr, cv2.IMREAD_COLOR)


def add_annotation_to_image(img, text, write_on_top=True):
    """
    Add Annotation to an image

    :param img: Image
    :type img: numpy.array
    :param text: text string
    :type text: str
    :param write_on_top: if you write the text on top
    :type write_on_top: bool
    :return: img with text written on it
    :rtype: numpy.array
    """
    font_scale_used = font_scale_normal
    thickness_used = thickness_small
    h_im, w_im = img.shape[:2]
    (text_width, text_height), baseline = cv2.getTextSize(text, font, fontScale=font_scale_used,
                                                          thickness=thickness_used)
    text_true_height = text_height + baseline
    if write_on_top:
        cv2.rectangle(img, (0, 0),
                      (int(text_width * 1.1), int(text_true_height * 1.35)),
                      WHITE, cv2.FILLED)
        cv2.putText(img, text,
                    (int(text_width * 0.05), int(text_true_height * 1.2 - baseline)),
                    font, font_scale_used, ORANGE, thickness_used)
    else:
        cv2.rectangle(img,
                      (int(text_width * 1.1), h_im - int(text_true_height * 1.35)),
                      (0, h_im),
                      WHITE, cv2.FILLED)
        cv2.putText(img, text,
                    (int(text_width * 0.05), h_im - baseline),
                    font, font_scale_used, ORANGE, thickness_used)
    return img


def undistort_image(img, mtx, dist):
    """
    Use camera intrinsics to undistort raw image

    :param img: Raw Image
    :type img: numpy.array
    :param mtx: Camera Intrinsics matrix
    :type mtx: list[list[float]]
    :param dist: Distortion Coefficient
    :type dist: list[list[float]]
    :return: Undistorted image
    :rtype: numpy.array
    """
    return cv2.undistort(src=img, cameraMatrix=mtx, distCoeffs=dist)


def resize_img(img, width=None, height=None, inter=cv2.INTER_AREA):
    """
    Resize an image. The user should precise only width or height if he wants to keep image's ratio

    :param img: OpenCV Image
    :type img: numpy.array
    :param width: Target Width
    :type width: int
    :param height: Target Height
    :type height: int
    :param inter: OpenCV interpolation flag
    :type inter: long
    :return: resized image
    :rtype: numpy.array
    """
    if width is None and height is None:
        return img

    height_init, width_init = img.shape[:2]

    if width is None:
        ratio = height / float(height_init)
        dim = (int(width_init * ratio), height)

    elif height is None:
        ratio = width / float(width_init)
        dim = (width, int(height_init * ratio))
    else:
        dim = (width, height)

    resized = cv2.resize(img, dim, interpolation=inter)

    return resized


def concat_imgs(tuple_imgs, axis=1):
    """
    Concat multiple images along 1 axis

    :param tuple_imgs: tuple of images
    :type tuple_imgs: tuple[numpy.array]
    :param axis: 0 means vertically and 1 means horizontally
    :type axis: int
    :return: Concat image
    :rtype: numpy.array
    """
    new_list_imgs = []
    for image in tuple_imgs:
        if len(image.shape) == 2:
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
        new_list_imgs.append(image)

    concat_im = np.concatenate(tuple(new_list_imgs), axis=axis)
    return concat_im


# ROS

def extract_img_from_ros_msg(ros_msg):
    """
    Take a ROS CompressedImage message and return the image uncompressed

    :param ros_msg: a ROS CompressedImage
    :type ros_msg: :sensor_msgs:`CompressedImage`
    :return: image uncompressed
    :rtype: numpy.array
    """
    data_image = ros_msg.data
    if not data_image:
        return None
    image = uncompress_image(data_image)
    return image


# DEBUG

def debug_threshold_color(img, color_hsv):
    """
    Return masked image to see the effect of color threshold

    :param img: OpenCV image
    :type img: numpy.array
    :param color_hsv: Color used for debug
    :type color_hsv: ColorHSV
    :return: Masked image
    :rtype: numpy.array
    """
    param_thresh = color_hsv.value
    thresh_img = threshold_hsv(img, *param_thresh)
    im_morph = morphological_transformations(thresh_img, MorphoType.OPEN, kernel_shape=(7, 7))
    return cv2.bitwise_and(img, img, mask=im_morph)


def debug_markers(img, workspace_ratio=1.0):
    """
    Display detected markers on an image

    :param img: OpenCV image which contain Niryo's markers
    :type img: numpy.array
    :param workspace_ratio: Ratio between the width and the height of the area represented by the markers
    :type workspace_ratio: float
    :return: (status, annotated image)
    :rtype: numpy.array
    """
    return draw_markers(img, workspace_ratio)



"""
This script shows 2 ways of making a vision pick and place.
You will need one workspace recorded in the Niryo studio
"""

# Imports
from pyniryo import *

# -- MUST Change these variables
simulation_mode = True
tool_used = ToolID.GRIPPER_1  # Tool used for picking
workspace_name = "workspace_1"  # Robot's Workspace Name
# Set robot address
robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simulation = "127.0.0.1"
robot_ip_address = robot_ip_address_simulation if simulation_mode else robot_ip_address_rpi

# -- Should Change these variables
# The pose from where the image processing happens
observation_pose = PoseObject(
    x=0.18, y=0.0, z=0.35,
    roll=0.0, pitch=1.57, yaw=-0.2,
)

# Center of the conditioning area
place_pose = PoseObject(
    x=0.0, y=-0.23, z=0.12,
    roll=0.0, pitch=1.57, yaw=-1.57
)


# -- MAIN PROGRAM

def vision_pick_n_place_1(niyro_robot):
    # Loop
    try_without_success = 0
    while try_without_success < 5:
        # Moving to observation pose
        niyro_robot.move_pose(*observation_pose.to_list())
        # Trying to get target pose using camera
        ret = niyro_robot.get_target_pose_from_cam(workspace_name,
                                                   height_offset=0.0,
                                                   shape=ObjectShape.ANY,
                                                   color=ObjectColor.ANY)
        obj_found, obj_pose, shape, color = ret
        if not obj_found:
            try_without_success += 1
            continue

        # At this point, we can do specific operation on the obj_pose if we want
        # to catch the object with a tilted angle for instance
        # ---
        # ---

        # Everything is good, so we going to pick the object
        niyro_robot.pick_from_pose(obj_pose)

        # Going to place
        niyro_robot.place_from_pose(place_pose)
        break


def vision_pick_n_place_2(niyro_robot):
    # Loop
    try_without_success = 0
    while try_without_success < 5:
        # Moving to observation pose
        niyro_robot.move_pose(*observation_pose.to_list())
        # Trying to pick target using camera
        ret = niyro_robot.vision_pick(workspace_name,
                                      height_offset=0.0,
                                      shape=ObjectShape.ANY,
                                      color=ObjectColor.ANY)
        obj_found, shape_ret, color_ret = ret
        if not obj_found:
            try_without_success += 1
            continue
        # Vision pick has succeed which means that Niryo One should have already catch the object !

        # Everything is good, so we going to place the object
        niyro_robot.place_from_pose(place_pose)
        break


if __name__ == '__main__':
    # Connect to robot
    client = NiryoRobot(robot_ip_address)
    # Changing tool
    client.update_tool()
    # Calibrate robot if robot needs calibration
    client.calibrate(CalibrateMode.AUTO)
    # Launching main process
    print("Starting Version 1")
    vision_pick_n_place_1(client)
    print("Starting Version 2")
    vision_pick_n_place_2(client)
    # Ending
    client.go_to_sleep()
    # Releasing connection
    client.close_connection()



import cv2
import numpy as np


class CalibrationObject:
    """
    Class containing intrinsic parameters of the camera used
    """

    def __init__(self, mtx, dist):
        self.__mtx = mtx
        self.__dist = dist

        self.__fx = self.__mtx[0][0]
        self.__fy = self.__mtx[1][1]
        self.__cx = self.__mtx[0][2]
        self.__cy = self.__mtx[1][2]

    @classmethod
    def set_from_values(cls, mtx, dist):
        return cls(mtx, dist)

    @classmethod
    def set_from_yaml(cls, yaml_file):
        mtx = np.reshape(yaml_file["mtx"], (3, 3))
        dist = np.expand_dims(yaml_file["dist"], axis=0)
        return cls(mtx, dist)

    @classmethod
    def set_empty(cls):
        mtx = np.zeros((3, 3), dtype=np.float)
        dist = []
        return cls(mtx, dist)

    def __str__(self):
        big_string = "mtx\n" + str(self.__mtx) + "\n"
        big_string += "dist\n" + str(self.__dist) + "\n"
        return big_string

    def get_center_position(self):
        return tuple([self.__cx, self.__cy])

    def get_center_position_int(self):
        return tuple([int(round(self.__cx)), int(round(self.__cy))])

    def get_intrinsic_parameters(self):
        return {
            "fx": self.__fx,
            "fy": self.__fy,
            "cx": self.__cx,
            "cy": self.__cy,
        }

    def get_cam_mtx_and_dist_coefs(self):
        return self.__mtx, self.__dist

    def undistort_image(self, img):
        return cv2.undistort(src=img, cameraMatrix=self.__mtx,
                             distCoeffs=self.__dist, newCameraMatrix=None)

    def meters_to_pixels(self, length, width, z_offset):
        u = (length / z_offset) * self.__fx
        v = (width / z_offset) * self.__fy

        return tuple([u, v])

    def get_camera_info(self):
        """
        Return value for ROS CameraInfo message
        """
        return self.__mtx, self.__dist

    def is_set(self):
        return self.__fx != 0.0 and self.__fy != 0.0 and self.__cx != 0.0 and self.__cy != 0.0



# Imports
from pyniryo import *

# - Constants
workspace_name = "workspace_1"  # Robot's Workspace Name
robot_ip_address = "x.x.x.x"

# The pose from where the image processing happens
observation_pose = PoseObject(
    x=0.16, y=0.0, z=0.35,
    roll=0.0, pitch=1.57, yaw=0.0,
)
# Place pose
place_pose = PoseObject(
    x=0.0, y=-0.2, z=0.12,
    roll=0.0, pitch=1.57, yaw=-1.57
)

# - Initialization

# Connect to robot
robot = NiryoRobot(robot_ip_address)
# Calibrate robot if robot needs calibration
robot.calibrate_auto()
# Updating tool
robot.update_tool()

# Initializing variables
offset_size = 0.05
max_catch_count = 4
shape_expected = ObjectShape.CIRCLE
color_expected = ObjectColor.RED

conveyor_id = robot.set_conveyor()

catch_count = 0
while catch_count < max_catch_count:
    # Turning conveyor on
    robot.run_conveyor(conveyor_id)
    # Moving to observation pose
    robot.move_pose(observation_pose)
    # Check if object is in the workspace
    obj_found, pos_array, shape, color = robot.detect_object(workspace_name,
                                                             shape=shape_expected,
                                                             color=color_expected)
    if not obj_found:
        robot.wait(0.5)  # Wait to let the conveyor turn a bit
        continue
    # Stopping conveyor
    robot.stop_conveyor(conveyor_id)
    # Making a vision pick
    obj_found, shape, color = robot.vision_pick(workspace_name,
                                                shape=shape_expected,
                                                color=color_expected)
    if not obj_found:  # If visual pick did not work
        continue

    # Calculate place pose and going to place the object
    next_place_pose = place_pose.copy_with_offsets(x_offset=catch_count * offset_size)
    robot.place_from_pose(next_place_pose)

    catch_count += 1

# Stopping conveyor
robot.stop_conveyor(conveyor_id)

# Unset conveyor
robot.unset_conveyor(conveyor_id)

robot.go_to_sleep()

robot.close_connection()



"""
Little Script which shows how to read a position from the TCP API
"""
# !/usr/bin/env python

from pyniryo import NiryoRobot

# Connecting to robot
niryo_robot = NiryoRobot(ip_address="192.168.1.15")

niryo_robot.calibrate_auto()

obj_pose = niryo_robot.get_target_pose_from_rel("workspace_1", height_offset=0.0,
                                                x_rel=0.2, y_rel=0.6, yaw_rel=1.57)

niryo_robot.move_pose(obj_pose)

# niryo_robot.set_learning_mode(True)

niryo_robot.close_connection()



# - Imports
from __future__ import print_function

# Python libraries
import time
import socket
import numpy as np
import sys

# Communication imports
from .enums_communication import *
from .communication_functions import dict_to_packet, receive_dict, receive_dict_w_payload

from .exceptions import *
from .objects import PoseObject, HardwareStatusObject, DigitalPinObject


class NiryoRobot(object):
    def __init__(self, ip_address=None):
        self.__ip_address = None
        self.__port = TCP_PORT
        self.__client_socket = None

        self.__timeout = TCP_TIMEOUT

        self.__is_running = True
        self.__is_connected = False

        # If user give IP Address, try to connect directly
        if ip_address is not None:
            self.connect(ip_address)

    def __del__(self):
        self.close_connection()

    def __str__(self):
        if self.__is_connected:
            msg = "\nConnected to server ({}) on port: {}\n".format(self.__ip_address, self.__port)
        else:
            msg = "Not Connected"

        return msg

    def __repr__(self):
        return self.__str__()

    # -- Connection
    def connect(self, ip_address):
        """
        Connect to the TCP Server

        :param ip_address: IP Address
        :type ip_address: str
        :rtype: None
        """
        self.__client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__client_socket.settimeout(self.__timeout)
        try:
            self.__client_socket.connect((ip_address, self.__port))
        except (socket.timeout, socket.error) as e:
            self.__client_socket = None
            raise ClientNotConnectedException("Unable to connect to the robot : {}".format(e))

        self.__is_connected = True
        self.__client_socket.settimeout(None)
        self.__ip_address = ip_address
        print("\nConnected to server ({}) on port: {}\n".format(ip_address, self.__port))

    def __close_socket(self):
        if self.__client_socket is not None and self.__is_connected is True:
            try:
                self.__client_socket.shutdown(socket.SHUT_RDWR)
                self.__client_socket.close()
            except socket.error:
                pass
            self.__is_connected = False
            print("\nDisconnected from robot\n")

    def close_connection(self):
        """
        Close connection with robot

        :rtype: None
        """
        self.__is_running = False
        self.__close_socket()
        self.__client_socket = None

    # -- SEND & RECEIVE

    # - Sending phase
    def __send_command(self, command_type, *parameter_list):
        if self.__is_connected is False:
            raise ClientNotConnectedException()
        if self.__client_socket is not None:
            try:
                ret_dict = self.__build_dict(command_type, *parameter_list)
                packet = dict_to_packet(ret_dict)
                self.__client_socket.send(packet)
            except socket.error as e:
                print(e, file=sys.stderr)
                raise HostNotReachableException()

    @staticmethod
    def __build_dict(command_type, *parameter_list):
        command = command_type.name
        new_param_list = []
        for parameter in parameter_list:
            if parameter is None:
                new_param_list.append("None")
            elif isinstance(parameter, Enum):
                new_param_list.append(parameter.name)
            elif isinstance(parameter, bool):
                new_param_list.append(str(parameter).upper())
            else:
                new_param_list.append(parameter)

        return {"command": command, "param_list": new_param_list}

    # - Receiving

    def __receive_answer(self, with_payload):
        try:
            if not with_payload:
                received_dict = receive_dict(sckt=self.__client_socket)
                payload = None
            else:
                received_dict, payload = receive_dict_w_payload(sckt=self.__client_socket)
        except socket.error as e:
            print(e, file=sys.stderr)
            raise HostNotReachableException()
        if not received_dict:
            raise HostNotReachableException()
        answer_status = received_dict["status"]
        if answer_status not in ["OK", "KO"]:
            raise InvalidAnswerException(answer_status)
        if received_dict["status"] != "OK":
            raise NiryoOneException("Command KO : {}".format(received_dict["message"]))
        list_ret_param = received_dict["list_ret_param"]
        if len(list_ret_param) == 1:
            list_ret_param = list_ret_param[0]
        if not with_payload:
            return list_ret_param
        else:
            return list_ret_param, payload

    # - Wrapping functions
    def __send_n_receive(self, command_type, *parameter_list, **kwargs):
        with_payload = kwargs.get("with_payload", False)

        self.__send_command(command_type, *parameter_list)
        return self.__receive_answer(with_payload=with_payload)

    # Parameters checker
    def __check_enum_belonging(self, value, enum_):
        """
        Check if a value belong to an enum
        """
        if value not in enum_:
            self.__raise_exception_expected_choice([v for v in enum_], value)

    def __check_list_belonging(self, value, list_):
        """
        Check if a value belong to a list
        """
        if value not in list_:
            self.__raise_exception_expected_choice(list_, value)

    def __check_range_belonging(self, value, range_min, range_max):
        """
        Check if a value belong to a range
        """
        if not range_min <= value <= range_max:
            self.__raise_exception_expected_range(range_min, range_max, value)

    def __check_dict_belonging(self, value, dict_):
        """
        Check if a value belong to a dictionary
        """
        if value not in dict_.keys():
            self.__raise_exception_expected_choice(dict_.keys(), value)

    def __check_type(self, value, type_):
        if type(value) is not type_:
            self.__raise_exception_expected_type(type_.__name__, value)

    def __check_instance(self, value, type_):
        if not isinstance(value, type_):
            self.__raise_exception_expected_type(type_.__name__, value)

    def __check_list_type(self, values_list, type_):
        for value in values_list:
            self.__check_type(value, type_)

    def __map_list(self, list_, type_):
        """
        Try to map a list to another type (Very useful for list like joints
        which are acquired as string)
        """
        try:
            map_list = list(map(type_, list_))
            return map_list
        except ValueError:
            self.__raise_exception_expected_type(type_.__name__, list_)

    def __transform_to_type(self, value, type_):
        """
        Try to change value type to another
        """
        try:
            value = type_(value)
            return value
        except ValueError:
            self.__raise_exception_expected_type(type_.__name__, value)

    # Error Handlers
    def __raise_exception_expected_choice(self, expected_choice, given):
        raise TcpCommandException("Expected one of the following: {}.\nGiven: {}".format(expected_choice, given))

    def __raise_exception_expected_type(self, expected_type, given):
        raise TcpCommandException("Expected type: {}.\nGiven: {}".format(expected_type, given))

    def __raise_exception_expected_range(self, range_min, range_max, given):
        raise TcpCommandException(
            "Expected the following condition: {} <= value <= {}\nGiven: {}".format(range_min, range_max, given))

    def __raise_exception(self, message):
        raise TcpCommandException("Exception message : {}".format(message))

    # -- Useful functions
    def __args_pose_to_list(self, *args):
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, PoseObject):
                return arg.to_list()
            else:
                pose_list = arg
        else:
            pose_list = args

        pose_list_float = self.__map_list(pose_list, float)
        if len(pose_list_float) != 6:
            self.__raise_exception("A pose should contain 6 elements (x, y, z, roll, pitch, yaw)")
        return pose_list_float

    def __args_joints_to_list(self, *args):
        """
        Convert args into a list
        Either if args = (1.1,5.6,-6.7) or args = ([1.1,5.6,-6.7],) , the
        function will return (1.1,5.6,-6.7)

        :param args: Union[list, tuple]
        :return: list of float
        """
        if len(args) == 1:
            args = args[0]

        joints = self.__map_list(args, float)
        if len(joints) != 6:
            self.__raise_exception("The robot has 6 joints")

        return joints

    # --- Public functions --- #

    # - Main purpose

    def calibrate(self, calibrate_mode):
        """
        Calibrate (manually or automatically) motors. Automatic calibration will do nothing
        if motors are already calibrated

        :param calibrate_mode: Auto or Manual
        :type calibrate_mode: CalibrateMode
        :rtype: None
        """
        self.__check_enum_belonging(calibrate_mode, CalibrateMode)
        self.__send_n_receive(Command.CALIBRATE, calibrate_mode)

    def calibrate_auto(self):
        """
        Start a automatic motors calibration if motors are not calibrated yet

        :rtype: None
        """
        self.calibrate(CalibrateMode.AUTO)

    def need_calibration(self):
        """
        Return a bool indicating whereas the robot motors need to be calibrate

        :rtype: bool
        """
        hardware_status = self.get_hardware_status()
        return hardware_status.calibration_needed

    @property
    def learning_mode(self):
        return self.get_learning_mode()

    def get_learning_mode(self):
        """
        Get learning mode state

        :return: ``True`` if learning mode is on
        :rtype: bool
        """
        return eval(self.__send_n_receive(Command.GET_LEARNING_MODE))

    @learning_mode.setter
    def learning_mode(self, value):
        self.set_learning_mode(value)

    def set_learning_mode(self, enabled):
        """
        Set learning mode if param is ``True``, else turn it off

        :param enabled: ``True`` or ``False``
        :type enabled: bool
        :rtype: None
        """
        self.__check_type(enabled, bool)
        self.__send_n_receive(Command.SET_LEARNING_MODE, enabled)

    def set_arm_max_velocity(self, percentage_speed):
        """
        Limit arm max velocity to a percentage of its maximum velocity

        :param percentage_speed: Should be between 1 & 100
        :type percentage_speed: int
        :rtype: None
        """
        self.__check_range_belonging(percentage_speed, 1, 100)
        self.__send_n_receive(Command.SET_ARM_MAX_VELOCITY, percentage_speed)

    def set_jog_control(self, enabled):
        """
        Set jog control mode if param is True, else turn it off

        :param enabled: ``True`` or ``False``
        :type enabled: bool
        :rtype: None
        """
        self.__check_type(enabled, bool)
        self.__send_n_receive(Command.SET_JOG_CONTROL, enabled)

    @staticmethod
    def wait(duration):
        """
        Wait for a certain time

        :param duration: duration in seconds
        :type duration: float
        :rtype: None
        """
        time.sleep(duration)

    # - Joints/Pose

    @property
    def joints(self):
        return self.get_joints()

    def get_joints(self):
        """
        Get joints value in radians
        You can also use a getter ::

            joints = robot.get_joints()
            joints = robot.joints

        :return: List of joints value
        :rtype: list[float]
        """
        return self.__send_n_receive(Command.GET_JOINTS)

    @property
    def pose(self):
        return self.get_pose()

    def get_pose(self):
        """
        Get end effector link pose as [x, y, z, roll, pitch, yaw].
        x, y & z are expressed in meters / roll, pitch & yaw are expressed in radians
        You can also use a getter ::

            pose = robot.get_pose()
            pose = robot.pose

        :rtype: PoseObject
        """
        data = self.__send_n_receive(Command.GET_POSE)
        pose_array = self.__map_list(data, float)
        pose_object = PoseObject(*pose_array)
        return pose_object

    def get_pose_quat(self):
        """
        Get end effector link pose in Quaternion coordinates

        :return: Position and quaternion coordinates concatenated in a list : [x, y, z, qx, qy, qz, qw]
        :rtype: list[float]
        """
        data = self.__send_n_receive(Command.GET_POSE_QUAT)
        pose_array = self.__map_list(data, float)
        return pose_array

    @joints.setter
    def joints(self, *args):
        self.move_joints(*args)

    def move_joints(self, *args):
        """
        Move robot joints. Joints are expressed in radians.

        All lines of the next example realize the same operation: ::

            robot.joints = [0.2, 0.1, 0.3, 0.0, 0.5, 0.0]
            robot.move_joints([0.2, 0.1, 0.3, 0.0, 0.5, 0.0])
            robot.move_joints(0.2, 0.1, 0.3, 0.0, 0.5, 0.0)

        :param args: either 6 args (1 for each joints) or a list of 6 joints
        :type args: Union[list[float], tuple[float]]
        :rtype: None
        """
        joints = self.__args_joints_to_list(*args)
        self.__send_n_receive(Command.MOVE_JOINTS, *joints)

    @pose.setter
    def pose(self, *args):
        self.move_pose(*args)

    def move_pose(self, *args):
        """
        Move robot end effector pose to a (x, y, z, roll, pitch, yaw) pose.
        x, y & z are expressed in meters / roll, pitch & yaw are expressed in radians

        All lines of the next example realize the same operation: ::

            robot.pose = [0.2, 0.1, 0.3, 0.0, 0.5, 0.0]
            robot.move_pose([0.2, 0.1, 0.3, 0.0, 0.5, 0.0])
            robot.move_pose(0.2, 0.1, 0.3, 0.0, 0.5, 0.0)
            robot.move_pose(PoseObject(0.2, 0.1, 0.3, 0.0, 0.5, 0.0))

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a ``PoseObject``
        :type args: Union[tuple[float], list[float], PoseObject]

        :rtype: None
        """
        pose_list = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.MOVE_POSE, *pose_list)

    def shift_pose(self, axis, shift_value):
        """
        Shift robot end effector pose along one axis

        :param axis: Axis along which the robot is shifted
        :type axis: RobotAxis
        :param shift_value: In meter for X/Y/Z and radians for roll/pitch/yaw
        :type shift_value: float
        :rtype: None
        """
        self.__check_enum_belonging(axis, RobotAxis)

        shift_value = self.__transform_to_type(shift_value, float)
        self.__send_n_receive(Command.SHIFT_POSE, axis, shift_value)

    def jog_joints(self, *args):
        """
        Jog robot joints'.
        Jog corresponds to a shift without motion planning.
        Values are expressed in radians.

        :param args: either 6 args (1 for each joints) or a list of 6 joints offset
        :type args: Union[list[float], tuple[float]]
        :rtype: None
        """
        joints_offset = self.__args_joints_to_list(*args)
        self.__send_n_receive(Command.JOG_JOINTS, *joints_offset)

    def jog_pose(self, *args):
        """
        Jog robot end effector pose
        Jog corresponds to a shift without motion planning
        Arguments are [dx, dy, dz, d_roll, d_pitch, d_yaw]
        dx, dy & dz are expressed in meters / d_roll, d_pitch & d_yaw are expressed in radians

        :param args: either 6 args (1 for each coordinates) or a list of 6 offset
        :type args: Union[list[float], tuple[float]]
        :rtype: None
        """
        pose_offset = self.__args_joints_to_list(*args)
        self.__send_n_receive(Command.JOG_POSE, *pose_offset)

    def move_linear_pose(self, *args):
        """
        Move robot end effector pose to a (x, y, z, roll, pitch, yaw) pose in a linear way

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[tuple[float], list[float], PoseObject]
        :rtype: None
        """
        pose_list = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.MOVE_LINEAR_POSE, *pose_list)

    def move_to_home_pose(self):
        """
        Move to a position where the forearm lays on shoulder

        :rtype: None
        """
        self.move_joints(0.0, 0.3, -1.3, 0.0, 0.0, 0.0)

    def go_to_sleep(self):
        """
        Go to home pose and activate learning mode

        :rtype: None
        """
        self.move_to_home_pose()
        self.set_learning_mode(True)

    def forward_kinematics(self, *args):
        """
        Compute forward kinematics of a given joints configuration and give the
        associated spatial pose

        :param args: either 6 args (1 for each joints) or a list of 6 joints
        :type args: Union[list[float], tuple[float]]
        :rtype: PoseObject
        """
        joints = self.__args_joints_to_list(*args)
        data = self.__send_n_receive(Command.FORWARD_KINEMATICS, *joints)

        pose_array = self.__map_list(data, float)
        pose_object = PoseObject(*pose_array)
        return pose_object

    def inverse_kinematics(self, *args):
        """
        Compute inverse kinematics

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a ``PoseObject``
        :type args: Union[tuple[float], list[float], PoseObject]

        :return: List of joints value
        :rtype: list[float]
        """
        pose_list = self.__args_pose_to_list(*args)

        return self.__send_n_receive(Command.INVERSE_KINEMATICS, *pose_list)

    # - Saved Pose

    def get_pose_saved(self, pose_name):
        """
        Get pose saved in from Niryo One's memory
        
        :param pose_name: Pose name in robot's memory
        :type pose_name: str 
        :return: Pose associated to pose_name
        :rtype: PoseObject
        """
        self.__check_type(pose_name, str)

        pose_list = self.__send_n_receive(Command.GET_POSE_SAVED, pose_name)
        return PoseObject(*pose_list)

    def save_pose(self, pose_name, *args):
        """
        Save pose in robot's memory
        
        :type pose_name: str
        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None 
        """
        self.__check_type(pose_name, str)
        pose_list = self.__args_pose_to_list(*args)

        self.__send_n_receive(Command.SAVE_POSE, pose_name, *pose_list)

    def delete_pose(self, pose_name):
        """
        Delete pose from robot's memory

        :type pose_name: str
        :rtype: None
        """
        self.__check_type(pose_name, str)
        self.__send_n_receive(Command.DELETE_POSE, pose_name)

    def get_saved_pose_list(self):
        """
        Get list of poses' name saved in robot memory

        :rtype: list[str]
        """
        return self.__send_n_receive(Command.GET_SAVED_POSE_LIST)

    # - Pick/Place

    def pick_from_pose(self, *args):
        """
        Execute a picking from a pose.

        A picking is described as : \n
        | * going over the object
        | * going down until height = z
        | * grasping with tool
        | * going back over the object

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None
        """
        pose = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.PICK_FROM_POSE, *pose)

    def place_from_pose(self, *args):
        """
        Execute a placing from a position.
        
        A placing is described as : \n
        | * going over the place
        | * going down until height = z
        | * releasing the object with tool
        | * going back over the place

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None
        """
        pose = self.__args_pose_to_list(*args)
        self.__send_n_receive(Command.PLACE_FROM_POSE, *pose)

    def pick_and_place(self, pick_pose, place_pos, dist_smoothing=0.0):
        """
        Execute a pick then a place

        :param pick_pose: Pick Pose : [x, y, z, roll, pitch, yaw] or PoseObject
        :type pick_pose: Union[list[float], PoseObject]
        :param place_pos: Place Pose : [x, y, z, roll, pitch, yaw] or PoseObject
        :type place_pos: Union[list[float], PoseObject]
        :param dist_smoothing: Distance from waypoints before smoothing trajectory
        :type dist_smoothing: float
        :rtype: None
        """
        pick_pose_list = self.__args_pose_to_list(pick_pose)
        place_pos_list = self.__args_pose_to_list(place_pos)

        self.__send_n_receive(Command.PICK_AND_PLACE, pick_pose_list, place_pos_list, dist_smoothing)

    # - Trajectories
    def get_trajectory_saved(self, trajectory_name):
        """
        Get trajectory saved in Niryo One's memory

        :type trajectory_name: str
        :return: Trajectory
        :rtype: list[list[float]]
        """
        self.__check_type(trajectory_name, str)
        return self.__send_n_receive(Command.GET_TRAJECTORY_SAVED, trajectory_name)

    def execute_trajectory_saved(self, trajectory_name):
        """
        Execute trajectory from Niryo One's memory

        :type trajectory_name: str
        :rtype: None
        """
        self.__check_type(trajectory_name, str)
        self.__send_n_receive(Command.EXECUTE_TRAJECTORY_SAVED, trajectory_name)

    def execute_trajectory_from_poses(self, list_poses, dist_smoothing=0.0):
        """
        Execute trajectory from list of poses

        :param list_poses: List of [x,y,z,qx,qy,qz,qw]
        :type list_poses: list[list[float]]
        :param dist_smoothing: Distance from waypoints before smoothing trajectory
        :type dist_smoothing: float
        :rtype: None
        """
        for i, pose in enumerate(list_poses):
            if len(pose) != 7:
                self.__raise_exception("7 parameters expected in a pose [x,y,z,qx,qy,qz,qw], "
                                       "{} parameters given".format(len(pose)))
            list_poses[i] = self.__map_list(pose, float)

        self.__send_n_receive(Command.EXECUTE_TRAJECTORY_FROM_POSES, list_poses, dist_smoothing)

    def save_trajectory(self, trajectory_name, list_poses):
        """
        Save trajectory in robot memory

        :type trajectory_name: str
        :param list_poses: List of [x,y,z,qx,qy,qz,qw]
        :type list_poses: list[list[float]]
        :rtype: None
        """
        self.__check_type(trajectory_name, str)
        for i, pose in enumerate(list_poses):
            if len(pose) != 7:
                self.__raise_exception("7 parameters expected in a pose [x,y,z,qx,qy,qz,qw], "
                                       "{} parameters given".format(len(pose)))
            list_poses[i] = self.__map_list(pose, float)

        self.__send_n_receive(Command.SAVE_TRAJECTORY, trajectory_name, list_poses)

    def delete_trajectory(self, trajectory_name):
        """
        Delete trajectory from robot's memory

        :type trajectory_name: str
        :rtype: None
        """
        self.__check_type(trajectory_name, str)
        self.__send_n_receive(Command.DELETE_TRAJECTORY, trajectory_name)

    def get_saved_trajectory_list(self):
        """
        Get list of trajectories' name saved in robot memory

        :rtype: list[str]
        """
        return self.__send_n_receive(Command.GET_SAVED_TRAJECTORY_LIST)

    # -- Tools

    @property
    def tool(self):
        return self.get_current_tool_id()

    def get_current_tool_id(self):
        """
        Get equipped tool Id

        :rtype: ToolID
        """
        tool_id = self.__send_n_receive(Command.GET_CURRENT_TOOL_ID)
        return ToolID[tool_id]

    def update_tool(self):
        """
        Update equipped tool

        :rtype: None
        """
        self.__send_n_receive(Command.UPDATE_TOOL)

    def grasp_with_tool(self):
        """
        Grasp with tool
        | This action correspond to
        | - Close gripper for Grippers
        | - Pull Air for Vacuum pump
        | - Activate for Electromagnet

        :rtype: None
        """
        self.__send_n_receive(Command.GRASP_WITH_TOOL)

    def release_with_tool(self):
        """
        Release with tool
        | This action correspond to
        | - Open gripper for Grippers
        | - Push Air for Vacuum pump
        | - Deactivate for Electromagnet

        :rtype: None
        """
        self.__send_n_receive(Command.RELEASE_WITH_TOOL)

    # - Gripper
    def open_gripper(self, speed=500):
        """
        Open gripper associated to 'gripper_id' with a speed 'speed'

        :param speed: Between 100 & 1000
        :type speed: int
        :rtype: None
        """
        speed = self.__transform_to_type(speed, int)

        self.__send_n_receive(Command.OPEN_GRIPPER, speed)

    def close_gripper(self, speed=500):
        """
        Close gripper associated to 'gripper_id' with a speed 'speed'

        :param speed: Between 100 & 1000
        :type speed: int
        :rtype: None
        """
        speed = self.__transform_to_type(speed, int)

        self.__send_n_receive(Command.CLOSE_GRIPPER, speed)

    # - Vacuum
    def pull_air_vacuum_pump(self):
        """
        Pull air of vacuum pump

        :rtype: None
        """
        self.__send_n_receive(Command.PULL_AIR_VACUUM_PUMP)

    def push_air_vacuum_pump(self):
        """
        Push air of vacuum pump

        :rtype: None
        """
        self.__send_n_receive(Command.PUSH_AIR_VACUUM_PUMP)

    # - Electromagnet
    def setup_electromagnet(self, pin_id):
        """
        Setup electromagnet on pin

        :param pin_id:
        :type pin_id: PinID
        :rtype: None
        """
        self.__check_enum_belonging(pin_id, PinID)

        self.__send_n_receive(Command.SETUP_ELECTROMAGNET, pin_id)

    def activate_electromagnet(self, pin_id):
        """
        Activate electromagnet associated to electromagnet_id on pin_id

        :param pin_id:
        :type pin_id: PinID
        :rtype: None
        """
        self.__check_enum_belonging(pin_id, PinID)

        self.__send_n_receive(Command.ACTIVATE_ELECTROMAGNET, pin_id)

    def deactivate_electromagnet(self, pin_id):
        """
        Deactivate electromagnet associated to electromagnet_id on pin_id

        :param pin_id:
        :type pin_id: PinID
        :rtype: None
        """
        self.__check_enum_belonging(pin_id, PinID)

        self.__send_n_receive(Command.DEACTIVATE_ELECTROMAGNET, pin_id)

    # - Hardware

    def set_pin_mode(self, pin_id, pin_mode):
        """
        Set pin number pin_id to mode pin_mode

        :param pin_id:
        :type pin_id: PinID
        :param pin_mode:
        :type pin_mode: PinMode
        :rtype: None
        """
        self.__check_enum_belonging(pin_id, PinID)
        self.__check_enum_belonging(pin_mode, PinMode)

        self.__send_n_receive(Command.SET_PIN_MODE, pin_id, pin_mode)

    def digital_write(self, pin_id, digital_state):
        """
        Set pin_id state to digital_state

        :param pin_id:
        :type pin_id: PinID
        :param digital_state:
        :type digital_state: PinState
        :rtype: None
        """
        self.__check_enum_belonging(pin_id, PinID)
        self.__check_enum_belonging(digital_state, PinState)

        self.__send_n_receive(Command.DIGITAL_WRITE, pin_id, digital_state)

    def digital_read(self, pin_id):
        """
        Read pin number pin_id and return its state

        :param pin_id:
        :type pin_id: PinID
        :rtype: PinState
        """
        self.__check_enum_belonging(pin_id, PinID)

        state_id = self.__send_n_receive(Command.DIGITAL_READ, pin_id)
        return PinState[state_id]

    @property
    def hardware_status(self):
        return self.get_hardware_status()

    def get_hardware_status(self):
        """
        Get hardware status : Temperature, Hardware version, motors names & types ...

        :return: Infos contains in a HardwareStatusObject
        :rtype: HardwareStatusObject
        """
        data = self.__send_n_receive(Command.GET_HARDWARE_STATUS)

        rpi_temperature = float(data[0])
        hardware_version = data[1]
        connection_up = eval(data[2])
        error_message = data[3]
        calibration_needed = eval(data[4])
        calibration_in_progress = eval(data[5])

        motor_names = data[6]
        motor_types = data[7]
        temperatures = self.__map_list(data[8], float)
        voltages = self.__map_list(data[9], float)
        hardware_errors = data[10]

        hardware_status = HardwareStatusObject(rpi_temperature, hardware_version, connection_up, error_message,
                                               calibration_needed, calibration_in_progress,
                                               motor_names, motor_types,
                                               temperatures, voltages, hardware_errors)
        return hardware_status

    @property
    def digital_io_state(self):
        return self.get_digital_io_state()

    def get_digital_io_state(self):
        """
        Get Digital IO state : Names, modes, states

        :return: List of DigitalPinObject instance
        :rtype: list[DigitalPinObject]
        """
        data = self.__send_n_receive(Command.GET_DIGITAL_IO_STATE)
        digital_pin_array = []
        for pin_info in data:
            pin_id, name, mode, state = pin_info
            digital_pin_array.append(DigitalPinObject(pin_id, name, mode, state))
        return digital_pin_array

    # - Conveyor

    def set_conveyor(self):
        """
        Activate a new conveyor and return its ID

        :return : New conveyor ID
        :rtype: ConveyorID
        """
        conveyor_id_str = self.__send_n_receive(Command.SET_CONVEYOR)
        conveyor_id = ConveyorID[conveyor_id_str]

        # If new conveyor has been found
        if conveyor_id != ConveyorID.NONE:
            return conveyor_id

        connected_conveyors_id = self.get_connected_conveyors_id()
        if not connected_conveyors_id:
            print("No conveyor connected !", file=sys.stderr)
            return ConveyorID.NONE
        else:
            print("No new conveyor detected, returning last connected conveyor", file=sys.stderr)
            return connected_conveyors_id[-1]

    def unset_conveyor(self, conveyor_id):
        """
        Remove specific conveyor.

        :param conveyor_id: Basically, ConveyorID.ONE or ConveyorID.TWO
        :type conveyor_id: ConveyorID
        """
        self.__send_n_receive(Command.UNSET_CONVEYOR, conveyor_id)

    def run_conveyor(self, conveyor_id, speed=50, direction=ConveyorDirection.FORWARD):
        """
        Run conveyor at id 'conveyor_id'

        :param conveyor_id:
        :type conveyor_id: ConveyorID
        :param speed:
        :type speed: int
        :param direction:
        :type direction: ConveyorDirection
        :rtype: None
        """
        self.control_conveyor(conveyor_id, control_on=True, speed=speed, direction=direction)

    def stop_conveyor(self, conveyor_id):
        """
        Stop conveyor at id 'conveyor_id'

        :param conveyor_id:
        :type conveyor_id: ConveyorID
        :rtype: None
        """
        self.control_conveyor(conveyor_id, control_on=False, speed=50, direction=ConveyorDirection.FORWARD)

    def control_conveyor(self, conveyor_id, control_on, speed, direction):
        """
        Control conveyor at id 'conveyor_id'

        :param conveyor_id:
        :type conveyor_id: ConveyorID
        :param control_on:
        :type control_on: bool
        :param speed: New speed which is a percentage of maximum speed
        :type speed: int
        :param direction: Conveyor direction
        :type direction: ConveyorDirection
        :rtype: None
        """
        self.__check_enum_belonging(conveyor_id, ConveyorID)
        self.__check_type(control_on, bool)
        self.__transform_to_type(speed, bool)
        self.__check_enum_belonging(direction, ConveyorDirection)

        self.__send_n_receive(Command.CONTROL_CONVEYOR, conveyor_id, control_on, speed, direction)

    def get_connected_conveyors_id(self):
        """

        :return: List of the connected conveyors' ID
        :rtype: list[ConveyorID]
        """
        conveyors_list_str = self.__send_n_receive(Command.GET_CONNECTED_CONVEYORS_ID)
        conveyors_list = [ConveyorID[conveyor] for conveyor in conveyors_list_str]

        return conveyors_list

    # - Vision
    def get_img_compressed(self):
        """
        Get image from video stream in a compressed format. 
        Use ``uncompress_image`` from the vision package to uncompress it 

        :return: string containing a JPEG compressed image
        :rtype: str
        """
        _, img = self.__send_n_receive(Command.GET_IMAGE_COMPRESSED, with_payload=True)
        return img

    def get_target_pose_from_rel(self, workspace_name, height_offset, x_rel, y_rel, yaw_rel):
        """
        Given a pose (x_rel, y_rel, yaw_rel) relative to a workspace, this function
        returns the robot pose in which the current tool will be able to pick an object at this pose.
        
        The height_offset argument (in m) defines how high the tool will hover over the workspace. If height_offset = 0,
        the tool will nearly touch the workspace.

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param x_rel: x relative pose (between 0 and 1)
        :type x_rel: float
        :param y_rel: y relative pose (between 0 and 1)
        :type y_rel: float
        :param yaw_rel: Angle in radians
        :type yaw_rel: float

        :return: target_pose
        :rtype: PoseObject
        """
        self.__check_type(workspace_name, str)
        self.__check_range_belonging(x_rel, 0.0, 1.0)
        self.__check_range_belonging(y_rel, 0.0, 1.0)

        height_offset, x_rel, y_rel, yaw_rel = self.__map_list([height_offset, x_rel, y_rel, yaw_rel], float)
        pose_array = self.__send_n_receive(Command.GET_TARGET_POSE_FROM_REL,
                                           workspace_name, height_offset, x_rel, y_rel, yaw_rel)
        pose_object = PoseObject(*pose_array)
        return pose_object

    def get_target_pose_from_cam(self, workspace_name, height_offset=0.0, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        First detects the specified object using the camera and then returns the robot pose in which the object can
        be picked with the current tool

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_pose, object_shape, object_color
        :rtype: (bool, PoseObject, ObjectShape, ObjectColor)
        """
        self.__check_type(workspace_name, str)
        height_offset = self.__transform_to_type(height_offset, float)
        self.__check_enum_belonging(shape, ObjectShape)
        self.__check_enum_belonging(color, ObjectColor)

        data_array = self.__send_n_receive(Command.GET_TARGET_POSE_FROM_CAM,
                                           workspace_name, height_offset, shape, color)
        obj_found = eval(data_array[0])
        if obj_found:
            pose_object = PoseObject(*data_array[1])
            shape_ret = data_array[2]
            color_ret = data_array[3]
        else:
            pose_object = PoseObject(0, 0, 0, 0, 0, 0)
            shape_ret = "ANY"
            color_ret = 'ANY'
        return obj_found, pose_object, ObjectShape[shape_ret], ObjectColor[color_ret]

    def __move_with_vision(self, command, workspace_name, height_offset, shape, color):
        self.__check_type(workspace_name, str)
        height_offset = self.__transform_to_type(height_offset, float)
        self.__check_enum_belonging(shape, ObjectShape)
        self.__check_enum_belonging(color, ObjectColor)

        data_array = self.__send_n_receive(command, workspace_name, height_offset, shape, color)

        obj_found = eval(data_array[0])
        if obj_found is True:
            shape_ret = data_array[1]
            color_ret = data_array[2]
        else:
            shape_ret = "ANY"
            color_ret = 'ANY'
        return obj_found, ObjectShape[shape_ret], ObjectColor[color_ret]

    def vision_pick(self, workspace_name, height_offset=0.0, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        Picks the specified object from the workspace. This function has multiple phases: \n
        | 1. detect object using the camera 
        | 2. prepare the current tool for picking 
        | 3. approach the object 
        | 4. move down to the correct picking pose 
        | 5. actuate the current tool 
        | 6. lift the object

        Example::

            robot = NiryoOneClient(ip_address="x.x.x.x")
            robot.calibrate_auto()
            robot.move_pose(<observation_pose>)
            obj_found, shape_ret, color_ret = robot.vision_pick(<workspace_name>,
                                                                height_offset=0.0,
                                                                shape=ObjectShape.ANY,
                                                                color=ObjectColor.ANY)

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_shape, object_color
        :rtype: (bool, ObjectShape, ObjectColor)
        """
        return self.__move_with_vision(Command.VISION_PICK, workspace_name, height_offset, shape, color)

    def move_to_object(self, workspace_name, height_offset, shape, color):
        """
        Same as `get_target_pose_from_cam` but directly moves to this position

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_shape, object_color
        :rtype: (bool, ObjectShape, ObjectColor)
        """
        return self.__move_with_vision(Command.MOVE_TO_OBJECT, workspace_name, height_offset, shape, color)

    def detect_object(self, workspace_name, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        Detect object in workspace and return its pose and characteristics

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_pose, object_shape, object_color
        :rtype: (bool, PoseObject, str, str)
        """
        self.__check_type(workspace_name, str)
        self.__check_enum_belonging(shape, ObjectShape)
        self.__check_enum_belonging(color, ObjectColor)

        data_array = self.__send_n_receive(Command.DETECT_OBJECT, workspace_name, shape, color)
        obj_found = eval(data_array[0])
        if not obj_found:
            rel_pose_array = 3 * [0.0]
            shape = "ANY"
            color = "ANY"
        else:
            rel_pose_array = data_array[1:4]
            shape = data_array[4]
            color = data_array[5]

        return obj_found, rel_pose_array, ObjectShape[shape], ObjectColor[color]

    def get_camera_intrinsics(self):
        """
        Get calibration object: camera intrinsics, distortions coefficients

        :return: camera intrinsics, distortions coefficients
        :rtype: (list[list[float]], list[list[float]])
        """
        data = self.__send_n_receive(Command.GET_CAMERA_INTRINSICS)

        mtx = np.reshape(data[0], (3, 3))
        dist = np.expand_dims(data[1], axis=0)
        return mtx, dist

    # - Workspace
    def save_workspace_from_robot_poses(self, workspace_name, pose_origin, pose_2, pose_3, pose_4):
        """
        Save workspace by giving the poses of the robot to point its 4 corners
        with the calibration Tip. Corners should be in the good order.
        Markers' pose will be deduced from these poses

        Poses should be either a list [x, y, z, roll, pitch, yaw] or a PoseObject

        :param workspace_name: workspace name
        :type workspace_name: str
        :param pose_origin:
        :type pose_origin: Union[list[float], PoseObject]
        :param pose_2:
        :type pose_2: Union[list[float], PoseObject]
        :param pose_3:
        :type pose_3: Union[list[float], PoseObject]
        :param pose_4:
        :type pose_4: Union[list[float], PoseObject]

        :rtype: None
        """
        self.__check_type(workspace_name, str)

        param_list = [workspace_name]
        for pose in (pose_origin, pose_2, pose_3, pose_4):
            pose_list = self.__args_pose_to_list(pose)
            param_list.append(pose_list)
        self.__send_n_receive(Command.SAVE_WORKSPACE_FROM_POSES, *param_list)

    def save_workspace_from_points(self, workspace_name, point_origin, point_2, point_3, point_4):
        """
        Save workspace by giving the points of worskpace's 4 corners. Points are written as [x, y, z]
        Corners should be in the good order.

        :param workspace_name: workspace name
        :type workspace_name: str
        :param point_origin:
        :type point_origin: list[float]
        :param point_2:
        :type point_2: list[float]
        :param point_3:
        :type point_3: list[float]
        :param point_4:
        :type point_4: list[float]
        :rtype: None
        """
        self.__check_type(workspace_name, str)

        param_list = [workspace_name]
        for point in (point_origin, point_2, point_3, point_4):
            param_list.append(self.__map_list(point, float))
        self.__send_n_receive(Command.SAVE_WORKSPACE_FROM_POINTS, *param_list)

    def delete_workspace(self, workspace_name):
        """
        Delete workspace from robot's memory

        :param workspace_name:
        :type workspace_name: str
        :rtype: None
        """
        self.__check_type(workspace_name, str)

        self.__send_n_receive(Command.DELETE_WORKSPACE, workspace_name)

    def get_workspace_poses(self, workspace_name):
        raise NotImplementedError

    def get_workspace_ratio(self, workspace_name):
        """
        Get workspace ratio from robot's memory

        :param workspace_name:
        :type workspace_name: str
        :rtype: float
        """
        self.__check_type(workspace_name, str)
        return self.__send_n_receive(Command.GET_WORKSPACE_RATIO, workspace_name)

    def get_workspace_list(self):
        """
        Get list of workspaces' name store in robot's memory

        :rtype: list[str]
        """
        return self.__send_n_receive(Command.GET_WORKSPACE_LIST)



"""
This script shows how to use kinematics with PyNiryo
"""

from pyniryo import NiryoRobot

simulation_mode = True
robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simulation = "127.0.0.1"

robot_ip_address = robot_ip_address_simulation if simulation_mode else robot_ip_address_rpi

# Connecting to robot
niyro_robot = NiryoRobot(robot_ip_address)

# Calibrating Robot
niyro_robot.calibrate_auto()

# Getting initial pose
initial_pose = niyro_robot.get_pose()

# Calculating pose with FK and moving to this pose
pose_target = niyro_robot.forward_kinematics(0.2, 0.0, -0.4, 0.0, 0.0, 0.0)
niyro_robot.move_pose(pose_target)

# Calculating joints via IK and moving to these joints
joints_target = niyro_robot.inverse_kinematics(initial_pose)
niyro_robot.move_joints(*joints_target)

# Leaving
niyro_robot.set_learning_mode(True)
niyro_robot.close_connection()



"""
This script allows to capture Niryo One's video streaming and to make some image processing on it
"""

# Imports
from pyniryo.api import *
import pyniryo.vision as vision

simulation_mode = True
# Set robot address
robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simulation = "127.0.0.1"
robot_ip_address = robot_ip_address_simulation if simulation_mode else robot_ip_address_rpi

# Set Observation Pose. It's where the robot will be placed for streaming
observation_pose = PoseObject(
    x=0.2, y=0.0, z=0.34,
    roll=0, pitch=1.57, yaw=-0.2,
)


def video_stream(niyro_robot):
    # Getting calibration param
    mtx, dist = niyro_robot.get_camera_intrinsics()
    # Moving to observation pose
    niyro_robot.move_pose(*observation_pose.to_list())

    while "User do not press Escape neither Q":
        # Getting image
        img_compressed = niyro_robot.get_img_compressed()
        # Uncompressing image
        img_raw = vision.uncompress_image(img_compressed)
        # Undistorting
        img_undistort = vision.undistort_image(img_raw, mtx, dist)
        # Trying to find markers
        workspace_found, res_img_markers = vision.debug_markers(img_undistort)
        # Trying to extract workspace if possible
        if workspace_found:
            img_workspace = vision.extract_img_workspace(img_undistort, workspace_ratio=1.0)
        else:
            img_workspace = None

        # - Display
        # Concatenating raw image and undistorted image
        concat_ims = vision.concat_imgs((img_raw, img_undistort))
        # Concatenating extracted workspace with markers annotation
        if img_workspace is not None:
            resized_img_workspace = vision.resize_img(img_workspace, height=res_img_markers.shape[0])
            res_img_markers = vision.concat_imgs((res_img_markers, resized_img_workspace))

        # Showing images
        vision.show_img("Images raw & undistorted", concat_ims, wait_ms=0)
        key = vision.show_img("Markers", res_img_markers, wait_ms=30)
        if key in [27, ord("q")]:  # Will break loop if the user press Escape or Q
            break

    niyro_robot.set_learning_mode(True)


if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Calibrate robot if robot needs calibration
    robot.calibrate_auto()
    # Launching main process
    video_stream(robot)
    # Releasing connection
    robot.close_connection()




"""
This script shows few functions you can use with the TCP API
"""

from pyniryo import *

simulation_mode = True

robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simulation = "127.0.0.1"

robot_ip_address = robot_ip_address_simulation if simulation_mode else robot_ip_address_rpi

# Connecting to robot
niryo_robot = NiryoRobot(robot_ip_address)

# Trying to calibrate
niryo_robot.calibrate(CalibrateMode.AUTO)
niryo_robot.set_learning_mode(False)

# Getting Joints
initial_joints = niryo_robot.get_joints()
print("Initial joints :\n" + str(initial_joints))

# Move Pose
niryo_robot.move_pose(0.2, 0.0, 0.25, 0.0, 0.0, 0.0)

# Shift pose
niryo_robot.shift_pose(RobotAxis.Y, 0.15)

# Going back to initial jointss
niryo_robot.move_joints(initial_joints)

niryo_robot.set_learning_mode(True)

if not simulation_mode:
    # Getting hardware information
    digital_pin_array = niryo_robot.get_digital_io_state()
    for digital_pin in digital_pin_array:
        print(digital_pin)

    hardware_data = niryo_robot.get_hardware_status()
    print(hardware_data)

# Turning learning mode ON
niryo_robot.set_learning_mode(True)

niryo_robot.close_connection()



from __future__ import print_function

import sys
from threading import Lock, Thread

import cv2
import time
from pyniryo.vision.image_functions import show_img_and_check_close


class WebcamVideoStream:
    def __init__(self, src=0,
                 name="WebcamVideoStream", display=False,
                 nb_threads=None):
        self.name = name
        self.display = display
        self._frame = None

        self._lock = Lock()
        self._thread = None

        self._stopped = False
        self._should_stop = False
        self.timestamp_last_frame = time.time()

        self.last_read_timestamp = self.timestamp_last_frame
        self.fps_list = [0.0 for _ in range(30)]
        self._failed_read = 0

        self._stream = cv2.VideoCapture(src)
        if self._stream is None or not self._stream.isOpened():
            sys.exit(4)

        (self._grabbed, self._frame) = self._stream.read()
        # logger.debug("Camera Object created and launched")
        if nb_threads is not None:
            cv2.setNumThreads(nb_threads)
        # logger.info("OpenCV is using {} threads".format(cv2.getNumThreads()))

    def start(self):
        if self._thread is not None and self._thread.isAlive():
            print("Thread already started", file=sys.stderr)
            return
        self._thread = Thread(target=self.update, name=self.name, args=())
        self._thread.daemon = True
        self._thread.start()
        return self

    def _get_new_timestamp(self):
        self.timestamp_last_frame = time.time()

    def read(self):
        with self._lock:
            return self._frame.copy()

    def update(self):
        while not self._should_stop:
            (self._grabbed, frame_obtain) = self._stream.read()
            if not self._grabbed:
                self._failed_read += 1
                print("Video Capture Read Nothing", file=sys.stderr)
                if self._failed_read > 30:
                    self._should_stop = True
                    cv2.destroyWindow(self.name)
                continue
            self._get_new_timestamp()
            self._failed_read = 0
            with self._lock:
                self._frame = frame_obtain.copy()
            if self.display and show_img_and_check_close(self.name, self._frame):
                self._should_stop = True
                cv2.destroyWindow(self.name)
            time.sleep(max(0., 30E-03 - (time.time() - self.timestamp_last_frame)))
        self._stopped = True

    def stop(self):
        print("Stopping Webcam's Stream", file=sys.stderr)

        # del self.stream
        self._should_stop = True



# Imports
from pyniryo import *

# Connecting to robot
niryo_robot = NiryoRobot("10.10.10.10")  # =< Replace by robot ip address

# Calibrating Robot
niryo_robot.calibrate_auto()
# Updating tool
niryo_robot.update_tool()

# Looping Forever
while True:
    # Moving to observation pause
    niryo_robot.move_pose(0.16, 0.0, 0.35, 0.0, 1.57, 0.0)

    # -- 1 -- ##
    img_compressed = niryo_robot.get_img_compressed()
    img = uncompress_image(img_compressed)
    if show_img_and_check_close("Image", img):
        break
    # // 1 // ##
    # -- 2 -- ##
    im_work = extract_img_workspace(img, workspace_ratio=1.0)
    if im_work is None:
        print("Unable to find markers")
        continue
    if show_img_and_check_close("Image Work", im_work):
        break
    # // 2 // ##

    # -- 3 -- ##
    low_thresh_list, high_thresh_list = [90, 115, 75], [115, 255, 255]
    img_thresh = threshold_hsv(im_work, low_thresh_list, high_thresh_list)
    if show_img_and_check_close("Thresh", img_thresh):
        break

    # // 3 // ##

    # -- 4 & 5 -- ##
    im_morpho = morphological_transformations(img_thresh, morpho_type=MorphoType.OPEN,
                                              kernel_shape=(5, 5), kernel_type=KernelType.ELLIPSE)
    if show_img_and_check_close("Morpho", im_morpho):
        break
    # // 4 & 5 // ##

    # -- 6 -- #
    best_contour = biggest_contour_finder(im_morpho)
    if len(best_contour) == 0:
        print("No blob found")
        continue
    show_img("Biggest Contour", draw_contours(im_morpho, [best_contour]))
    # // 6 // #
    # -- 7 & 8 -- #
    cx, cy = get_contour_barycenter(best_contour)
    height, width = img_thresh.shape
    x_rel = float(cx) / width
    y_rel = float(cy) / height
    angle = get_contour_angle(best_contour)
    # // 7 & 8// #
    # -- 9 -- #
    ws_name = "workspace_1"
    height_offset = 0.0
    obj_pose = niryo_robot.get_target_pose_from_rel(ws_name, height_offset,
                                                    x_rel, y_rel, angle)
    # // 9 // #

    # -- 10 -- #
    place_pose = PoseObject(
        x=0.0, y=0.20, z=0.3,
        roll=0.0, pitch=1.57, yaw=1.57
    )

    # Picking the object
    niryo_robot.pick_from_pose(obj_pose)

    # Placing the object
    niryo_robot.place_from_pose(place_pose)
    # // 10 // #

niryo_robot.go_to_sleep()
niryo_robot.close_connection()
